package it.amattioli.dominate.hibernate;

import it.amattioli.dominate.Entity;
import it.amattioli.dominate.specifications.hibernate.CriteriaAssembler;
import it.amattioli.dominate.specifications.hibernate.HqlAssembler;
import it.amattioli.dominate.util.GenericComparator;
import it.amattioli.dominate.Filter;
import it.amattioli.dominate.Repository;
import it.amattioli.dominate.Specification;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;
import org.hibernate.Criteria;
import org.hibernate.Query;
import org.hibernate.criterion.Order;

/**
 * A repository to access all the entities of a certain class persisted
 * with Hibernate.
 * 
 * @param <T> The class of the objects that can be accessed using this repository
 * @param <I> The class of the identifier of the objects that can be accessed using this repository
 * @author a.mattioli
 *
 */
public class ClassHibernateRepository<I extends Serializable, T extends Entity<I>> extends AbstractHibernateRepository<I, T> {
    private Class<T> repositoryClass;
    private Filter lastCriteria;
    private boolean dbOrder = true;
    private boolean cacheable = false;

    /**
     * Create a repository for the entities whose class is passed as parameter
     * 
     * @param repositoryClass the class of the entities that can be accessed
     *        using this repository
     */
    public ClassHibernateRepository(final Class<T> repositoryClass) {
        this.repositoryClass = repositoryClass;
    }

    /**
     * Retrieves the class of the entities that can be accessed using this repository
     *
     * @return the class of the entities that can be accessed using this repository
     */
    public Class<T> getRepositoryClass() {
        return repositoryClass;
    }

    @SuppressWarnings("unchecked")
    public T get(final I id) {
        return (T) getHibernateSession().get(getRepositoryClass(), id);
    }

    public void put(final T object) {
        getHibernateSession().saveOrUpdate(object);
    }

    public void refresh(final I objectId) {
        refresh(get(objectId));
    }

    public void refresh(final T object) {
        getHibernateSession().evict(object);
    }

    @Override
	public boolean isRemoveAllowed() {
		return true;
	}

	@Override
	public void remove(I objectId) {
		remove(get(objectId));
	}

	@Override
	public void remove(T object) {
		getHibernateSession().delete(object);
	}

	protected org.hibernate.Criteria defaultCriteria() {
        return getHibernateSession().createCriteria(getRepositoryClass());
    }

    public boolean isDbOrder() {
		return dbOrder;
	}

	public void setDbOrder(boolean dbOrder) {
		this.dbOrder = dbOrder;
	}
	
	protected String addOrder(String hqlQuery) {
		if (isDbOrder() && getOrderProperty() != null) {
		    hqlQuery = hqlQuery + " " + getHqlOrder();
		}
		return hqlQuery;
	}
    
    protected void addOrder(final org.hibernate.Criteria crit) {
        if (isDbOrder()) {
        	for (it.amattioli.dominate.repositories.Order order: getOrders()) {
	            if (order.isReverse()) {
	                crit.addOrder(Order.desc(order.getProperty()));
	            } else {
	                crit.addOrder(Order.asc(order.getProperty()));
	            }
        	}
        }
    }
    
    protected void inMemoryOrder(List<T> result) {
    	if (!isDbOrder() && getOrderProperty() != null) {
    		Collections.sort(result,new GenericComparator<T>(getOrderProperty()));
            if (isReverseOrder()) {
                Collections.reverse(result);
            }
    	}
    }
    
    @SuppressWarnings("unchecked")
    public List<T> list() {
        lastCriteria = null;
        org.hibernate.Criteria crit = defaultCriteria();
        addOrder(crit);
        setCriteriaLimits(crit);
        crit.setCacheable(cacheable);
        List<T> result = crit.list();
        inMemoryOrder(result);
        return result;
    }

    @Deprecated
    @SuppressWarnings("unchecked")
    public List<T> list(final Filter criteria) {
        lastCriteria = criteria;
        List<T> result;
        HibernateFilter hCriteria = (HibernateFilter) criteria;
        String hqlQuery = hCriteria.getHqlQuery();
        Predicate additional = hCriteria.getCollectionPredicate();
        int first = criteria.getFirst() == 0 ? getFirst() : criteria.getFirst();
        int last = criteria.getLast() == 0 ? getLast() : criteria.getLast();
        if (hqlQuery != null) {
            hqlQuery = addOrder(hqlQuery);
            Query query = getHibernateSession().createQuery(hqlQuery);
            hCriteria.setHqlParameters(query);
            if (additional == null) {
                if (first != 0) {
                    query.setFirstResult(first);
                }
                
                if (last != 0) {
                    query.setMaxResults(last);
                }
            }
            query.setCacheable(cacheable);
            result = query.list();
        } else {
            org.hibernate.Criteria crit = defaultCriteria();
            ((HibernateFilter) criteria).customize(crit);
            addOrder(crit);
            if (additional == null) {
            	if (first != 0) {
                    crit.setFirstResult(first);
                }
                if (last != 0) {
                    crit.setMaxResults(last);
                }
            }
            crit.setCacheable(cacheable);
            result = crit.list();
        }
        if (additional != null) {
            CollectionUtils.filter(result, additional);
            result = result.subList(first, last == 0 ? result.size() : Math.min(last,result.size()));
        }
        inMemoryOrder(result);
        return result;
    }
    
    private List<T> listByCriteria(Specification<T> spec) {
    	CriteriaAssembler cAssembler = new CriteriaAssembler(defaultCriteria());
    	if (!spec.supportsAssembler(cAssembler)) {
    		return null;
    	}
    	List<T> result;
    	spec.assembleQuery(cAssembler);
    	Criteria crit = cAssembler.assembledCriteria();
    	addOrder(crit);
    	if (!cAssembler.hasAdditionalPredicate()) {
        	setCriteriaLimits(crit);
    	}
        crit.setCacheable(cacheable);
        result = crit.list();
        result = applyAdditionalPredicate(result, cAssembler.additionalPredicate());
        inMemoryOrder(result);
		return result;
    }

	private void setCriteriaLimits(Criteria crit) {
		if (getFirst() != 0) {
		    crit.setFirstResult(getFirst());
		}
		if (getLast() != 0) {
		    crit.setMaxResults(getLast());
		}
	}

    private List<T> listByHql(Specification<T> spec) {
    	List<T> result;
		HqlAssembler hAssembler = new HqlAssembler("from " + getRepositoryClass().getName() + " ", getOrders());
    	if (!spec.supportsAssembler(hAssembler)) {
    		return null;
    	}
		spec.assembleQuery(hAssembler);
		Query query = hAssembler.assembledHqlQuery(getHibernateSession());
		if (!hAssembler.hasAdditionalPredicate()) {
    		setHqlLimits(query);
		}
        query.setCacheable(cacheable);
        result = query.list();
        result = applyAdditionalPredicate(result, hAssembler.additionalPredicate());
        inMemoryOrder(result);
		return result;
    }

	private void setHqlLimits(Query query) {
		if (getFirst() != 0) {
		    query.setFirstResult(getFirst());
		}
		if (getLast() != 0) {
		    query.setMaxResults(getLast());
		}
	}

    private List<T> applyAdditionalPredicate(List<T> result, Predicate additional) {
		if (additional != null) {
            CollectionUtils.filter(result, additional);
            result = result.subList(getFirst(), getLast() == 0 ? result.size() : Math.min(getLast(), result.size()));
        }
		return result;
	}
    
    @Override
	public List<T> list(Specification<T> spec) {
    	List<T> result = listByCriteria(spec);
    	if (result == null) {
    		result = listByHql(spec);
    	}
    	if (result == null) {
    		throw new IllegalArgumentException("Unsupported specification");
    	}
    	return result;
    	/*
    	List<T> result;
    	int first = getFirst();
        int last = getLast();
    	CriteriaAssembler cAssembler = new CriteriaAssembler(defaultCriteria());
    	Predicate additional = null;
        if (spec.supportsAssembler(cAssembler)) {
        	spec.assembleQuery(cAssembler);
        	Criteria crit = cAssembler.assembledCriteria();
        	addOrder(crit);
        	additional = cAssembler.additionalPredicate();
        	if (additional == null) {
            	if (first != 0) {
                    crit.setFirstResult(first);
                }
                if (last != 0) {
                    crit.setMaxResults(last);
                }
        	}
            crit.setCacheable(cacheable);
            result = crit.list();
            
        } else {
        	String hqlOrder = isDbOrder() && getOrderProperty() != null ? getOrderProperty() : "";
			HqlAssembler hAssembler = new HqlAssembler("from " + getRepositoryClass().getName() + " ", hqlOrder, isReverseOrder());
        	if (spec.supportsAssembler(hAssembler)) {
        		spec.assembleQuery(hAssembler);
        		Query query = hAssembler.assembledHqlQuery(getHibernateSession());
        		additional = hAssembler.additionalPredicate();
        		if (additional == null) {
            		if (first != 0) {
                        query.setFirstResult(first);
                    }
                    if (last != 0) {
                        query.setMaxResults(last);
                    }
        		}
                query.setCacheable(cacheable);
                result = query.list();
        	} else {
        		throw new IllegalArgumentException("Unsupported specification");
        	}
        }
        if (additional != null) {
            CollectionUtils.filter(result, additional);
            result = result.subList(first, last == 0 ? result.size() : last);
        }
        inMemoryOrder(result);
		return result;
		*/
	}

    @Deprecated
	public Filter getLastCriteria() {
        return lastCriteria;
    }

    public <J extends Serializable, E extends Entity<J>> Repository<J, E> getDetailRepository(Collection<E> details) {
        CollectionHibernateRepository<J, E> rep = new CollectionHibernateRepository<J, E>(details);
        rep.setSessionManager(getSessionManager());
        return rep;
    }

	public boolean isCacheable() {
		return cacheable;
	}

	public void setCacheable(boolean cacheable) {
		this.cacheable = cacheable;
	}
	
	@Deprecated
	public void fillCollection(Filter criteria, Collection<? super T>... toBeFilled) {
		//TODO: andrebbe implementato usando il metodo scroll di Query e Criteria
		Collection<T> result = list(criteria);
		for (T currRes: result) {
			for(Collection<? super T> currColl: toBeFilled) {
				currColl.add(currRes);
			}
		}
	}

	public void fillCollection(Collection<? super T>... toBeFilled) {
		//TODO: andrebbe implementato usando il metodo scroll di Query e Criteria
		Collection<T> result = list();
		for (T currRes: result) {
			for(Collection<? super T> currColl: toBeFilled) {
				currColl.add(currRes);
			}
		}
	}

	@Override
	public String toString() {
		return "ClassHibernateRepository[" + repositoryClass + "]";
	}
}
